import time
from multiprocessing import Process, Queue
from typing import Optional

from psutil import NoSuchProcess
from psutil import Process as PsutilProcess

from glasswall.multiprocessing.deletion import force_object_under_target_size
from glasswall.multiprocessing.tasks import Task, TaskResult, execute_task_and_put_in_queue
from glasswall.utils import round_up


def get_total_memory_in_gb(process: Process) -> float:
    total_memory: float = 0
    try:
        psutil_process = PsutilProcess(process.pid)
        total_memory = psutil_process.memory_info().rss

        # Include memory usage of child processes
        for child in psutil_process.children(recursive=True):
            total_memory += child.memory_info().rss

        return total_memory / 1024 ** 3  # Convert bytes to GB

    except NoSuchProcess:
        # Handle the case where the process does not exist
        return 0.0


class TaskWatcher:
    process: Process
    start_time: float
    end_time: float
    elapsed_time: float

    def __init__(
        self,
        task: Task,
        task_results_queue: "Queue[TaskResult]",
        timeout_seconds: Optional[float] = None,
        memory_limit_in_gb: Optional[float] = None,
        sleep_time: float = 0.001,
        sleep_time_memory_limit: float = 1,
        auto_start: bool = True,
    ):
        self.task = task
        self.task_results_queue = task_results_queue
        self.timeout_seconds = timeout_seconds
        self.memory_limit_in_gb = memory_limit_in_gb
        self.sleep_time = sleep_time
        self.sleep_time_memory_limit = sleep_time_memory_limit
        self.auto_start = auto_start

        self.watcher_queue: "Queue[TaskResult]" = Queue()

        self.exception = None
        self.timed_out: bool = False
        self.out_of_memory: bool = False

        if self.auto_start:
            self.start_task()
            self.watch_task()
            self.update_queue()

    def start_task(self) -> None:
        self.process = Process(
            target=execute_task_and_put_in_queue,
            args=(self.task, self.watcher_queue,)
        )
        self.process.start()
        self.start_time = time.time()

    def terminate_task(self) -> None:
        self.process.terminate()

    def terminate_task_with_timeout(self) -> None:
        self.terminate_task()
        self.timed_out = True
        self.exception = TimeoutError()

    def terminate_task_with_out_of_memory(self) -> None:
        self.terminate_task()
        self.out_of_memory = True
        self.exception = MemoryError()

    def watch_task(self) -> None:
        last_memory_limit_check = time.time()
        while self.process.is_alive():
            # Monitor for timeout exceeded
            if self.timeout_seconds:
                if time.time() - self.start_time > self.timeout_seconds:
                    self.terminate_task_with_timeout()
                    break

            # Monitor for memory limit exceeded
            if self.memory_limit_in_gb:
                if time.time() - last_memory_limit_check > self.sleep_time_memory_limit:
                    last_memory_limit_check = time.time()
                    memory_usage_in_gb = get_total_memory_in_gb(process=self.process)
                    if memory_usage_in_gb > self.memory_limit_in_gb:
                        self.terminate_task_with_out_of_memory()
                        break

            if self.sleep_time:
                time.sleep(self.sleep_time)

        self.end_time = time.time()
        self.elapsed_time = round_up(self.end_time - self.start_time, decimals=2)
        self.process.join()

    def update_queue(self) -> None:
        if self.exception:
            # TimeoutError or MemoryError
            task_result = TaskResult(
                self.task,
                success=False,
                exception=self.exception,
            )  # TODO add memoryusage
        else:
            task_result = self.watcher_queue.get()

        task_result.task = self.task
        task_result.timeout_seconds = self.timeout_seconds
        task_result.memory_limit_in_gb = self.memory_limit_in_gb

        task_result.start_time = self.start_time
        task_result.end_time = self.end_time
        task_result.elapsed_time = self.elapsed_time
        task_result.timed_out = self.timed_out
        task_result.out_of_memory = self.out_of_memory

        # Ensure task_result can fit in queue
        task_result = force_object_under_target_size(task_result, target_size=8192)

        self.task_results_queue.put(task_result)





import logging
import os
from datetime import datetime

import glasswall


# 2021-03-22 17:34:59.428	glasswall.config.logging     	INFO    	__init__		 Loaded Glasswall Editor version 2.0 from C:\Users\AngusRoberts\Documents\Azure\gated-check-in\gated_check_in\data\combined_libraries\gcin\glasswall_core2.dll
fmt = "%(asctime)s.%(msecs)03d\t%(name)-29s\t%(levelname)-8s\t%(funcName)s\t\t%(message)s"

# 2020-06-25 15:04:59
datefmt = "%Y-%m-%d %H:%M:%S"

log_formatter = logging.Formatter(fmt, datefmt)

# %temp%/glasswall/logs/2020-06-25 150459.txt
log_file_path = os.path.join(glasswall._TEMPDIR, "logs", f'{datetime.now().strftime("%Y-%m-%d %H%M%S")}.txt')

# Ensure log directory exists
os.makedirs(os.path.dirname(log_file_path), exist_ok=True)

# Logging to file
log = logging.getLogger(__name__)
log_handler = logging.FileHandler(log_file_path, mode="w", delay=True)
log_handler.setFormatter(log_formatter)
log.addHandler(log_handler)
if os.environ.get("glasswall_log_level"):
    log.setLevel(os.environ.get("glasswall_log_level"))
else:
    log.setLevel(logging.DEBUG)

# Logging to console
console = logging.StreamHandler()
console.setFormatter(log_formatter)
if os.environ.get("glasswall_console_level"):
    console.setLevel(os.environ.get("glasswall_console_level"))
else:
    console.setLevel(logging.INFO)
log.addHandler(console)


def format_object(obj):
    if not hasattr(obj, "__dict__"):
        return ""
    return "\n\t" + "\n\t".join(f"{k}: {v}" for k, v in obj.__dict__.items())





import ctypes as ct
import io
import os
from typing import Optional, Union

import glasswall
from glasswall import utils
from glasswall.config.logging import log
from glasswall.libraries.library import Library
from glasswall.libraries.word_search import errors, successes


class WordSearch(Library):
    """ A high level Python wrapper for Glasswall WordSearch. """

    def __init__(self, library_path: str):
        super().__init__(library_path=library_path)
        self.library = self.load_library(os.path.abspath(library_path))

        log.info(f"Loaded Glasswall {self.__class__.__name__} version {self.version()} from {self.library_path}")

    def version(self):
        """ Returns the Glasswall library version.

        Returns:
            version (str): The Glasswall library version.
        """
        # API function declaration
        self.library.GwWordSearchVersion.restype = ct.c_char_p

        # API call
        version = self.library.GwWordSearchVersion()

        # Convert to Python string
        version = ct.string_at(version).decode()

        return version

    @glasswall.utils.deprecated_alias(xml_config="content_management_policy")
    def redact_file(self, input_file: Union[str, bytes, bytearray, io.BytesIO], content_management_policy: Union[str, bytes, bytearray, io.BytesIO], output_file: Union[None, str] = None, output_report: Union[None, str] = None, homoglyphs: Union[None, str, bytes, bytearray, io.BytesIO] = None, raise_unsupported: bool = True):
        """ Redacts text from input_file using the given content_management_policy and homoglyphs file, optionally writing the redacted file and report to the paths specified by output_file and output_report.

        Args:
            input_file (Union[str, bytes, bytearray, io.BytesIO]): The input file path or bytes.
            content_management_policy (Union[str, bytes, bytearray, io.BytesIO)]): The content management policy to apply.
            output_file (Union[None, str], optional): Default None. If str, write output_file to that path.
            output_report (Union[None, str], optional): Default None. If str, write output_file to that path.
            homoglyphs (Union[None, str, bytes, bytearray, io.BytesIO)], optional): Default None. The homoglyphs json file path or bytes.
            raise_unsupported (bool, optional): Default True. Raise exceptions when Glasswall encounters an error. Fail silently if False.

        Returns:
            gw_return_object (glasswall.GwReturnObj): An instance of class glasswall.GwReturnObj containing attributes: "status" (int), "output_file" (bytes), "output_report" (bytes)
        """
        # Validate arg types
        if not isinstance(input_file, (str, bytes, bytearray, io.BytesIO)):
            raise TypeError(input_file)
        if not isinstance(content_management_policy, (str, bytes, bytearray, io.BytesIO, glasswall.content_management.policies.policy.Policy)):
            raise TypeError(content_management_policy)
        if not isinstance(output_file, (type(None), str)):
            raise TypeError(output_file)
        if not isinstance(output_report, (type(None), str)):
            raise TypeError(output_report)
        if not isinstance(homoglyphs, (type(None), str, bytes, bytearray, io.BytesIO)):
            raise TypeError(homoglyphs)

        # Convert string path arguments to absolute paths
        if isinstance(output_file, str):
            output_file = os.path.abspath(output_file)

        if isinstance(output_report, str):
            output_report = os.path.abspath(output_report)

        # Convert inputs to bytes
        if isinstance(input_file, str):
            with open(input_file, "rb") as f:
                input_file_bytes = f.read()
        elif isinstance(input_file, (bytes, bytearray, io.BytesIO)):
            input_file_bytes = utils.as_bytes(input_file)
        # warn if input_file is 0 bytes
        if not input_file_bytes:
            log.warning(f"input_file is 0 bytes\n\tinput_file: {input_file}")

        if isinstance(homoglyphs, str):
            with open(homoglyphs, "rb") as f:
                homoglyphs_bytes = f.read()
        elif isinstance(homoglyphs, (bytes, bytearray, io.BytesIO)):
            homoglyphs_bytes = utils.as_bytes(homoglyphs)
        elif isinstance(homoglyphs, type(None)):
            # Load default
            with open(os.path.join(glasswall._ROOT, "config", "word_search", "homoglyphs.json"), "rb") as f:
                homoglyphs_bytes = f.read()

        if isinstance(content_management_policy, str) and os.path.isfile(content_management_policy):
            with open(content_management_policy, "rb") as f:
                content_management_policy = f.read()
        content_management_policy = utils.validate_xml(content_management_policy)

        # Variable initialisation
        ct_input_buffer = ct.c_char_p(input_file_bytes)
        ct_input_buffer_length = ct.c_size_t(len(input_file_bytes))
        ct_output_buffer = ct.c_void_p()
        ct_output_buffer_length = ct.c_size_t()
        ct_output_report_buffer = ct.c_void_p()
        ct_output_report_buffer_length = ct.c_size_t()
        ct_homoglyphs = ct.c_char_p(homoglyphs_bytes)
        ct_content_management_policy = ct.c_char_p(content_management_policy.encode())
        gw_return_object = glasswall.GwReturnObj()

        with utils.CwdHandler(new_cwd=self.library_path):
            gw_return_object.status = self.library.GwWordSearch(
                ct_input_buffer,
                ct_input_buffer_length,
                ct.byref(ct_output_buffer),
                ct.byref(ct_output_buffer_length),
                ct.byref(ct_output_report_buffer),
                ct.byref(ct_output_report_buffer_length),
                ct_homoglyphs,
                ct_content_management_policy
            )

        gw_return_object.output_file = utils.buffer_to_bytes(
            ct_output_buffer,
            ct_output_buffer_length
        )
        gw_return_object.output_report = utils.buffer_to_bytes(
            ct_output_report_buffer,
            ct_output_report_buffer_length
        )

        input_file_repr = f"{type(input_file_bytes)} length {len(input_file_bytes)}" if not isinstance(input_file, str) else input_file
        output_file_repr = f"{type(gw_return_object.output_file)} length {len(gw_return_object.output_file)}"
        output_report_repr = f"{type(gw_return_object.output_report)} length {len(gw_return_object.output_report)}"
        homoglyphs_repr = f"{type(homoglyphs_bytes)} length {len(homoglyphs_bytes)}" if not isinstance(homoglyphs, str) else homoglyphs
        if gw_return_object.status not in successes.success_codes:
            log.error(f"\n\tinput_file: {input_file_repr}\n\toutput_file: {output_file_repr}\n\toutput_report: {output_report_repr}\n\thomoglyphs: {homoglyphs_repr}\n\tstatus: {gw_return_object.status}\n\tcontent_management_policy:\n{content_management_policy}")
            if raise_unsupported:
                raise errors.error_codes.get(gw_return_object.status, errors.UnknownErrorCode)(gw_return_object.status)
        else:
            log.debug(f"\n\tinput_file: {input_file_repr}\n\toutput_file: {output_file_repr}\n\toutput_report: {output_report_repr}\n\thomoglyphs: {homoglyphs_repr}\n\tstatus: {gw_return_object.status}\n\tcontent_management_policy:\n{content_management_policy}")

        # Write output file
        if gw_return_object.output_file:
            if isinstance(output_file, str):
                os.makedirs(os.path.dirname(output_file), exist_ok=True)
                with open(output_file, "wb") as f:
                    f.write(gw_return_object.output_file)

        # Write output report
        if gw_return_object.output_report:
            if isinstance(output_report, str):
                os.makedirs(os.path.dirname(output_report), exist_ok=True)
                with open(output_report, "wb") as f:
                    f.write(gw_return_object.output_report)

        if input_file_bytes and not gw_return_object.output_file:
            # input_file_bytes was not empty but output_file is unexpectedly empty
            log.error(f"output_file empty\n\tinput_file: {input_file_repr}\n\tct_output_buffer: {ct_output_buffer}\n\tct_output_buffer_length: {ct_output_buffer_length}\n\toutput_file: {gw_return_object.output_file}")
            if raise_unsupported:
                raise errors.WordSearchError(f"Unexpected empty output_file after calling GwWordSearch\n\toutput_file: {output_file}")

        if input_file_bytes and not gw_return_object.output_report:
            # input_file_bytes was not empty but output_report is unexpectedly empty
            log.error(f"output_report empty\n\tinput_file: {input_file_repr}\n\tct_output_report_buffer: {ct_output_report_buffer}\n\tct_output_report_buffer_length: {ct_output_report_buffer_length}")
            if raise_unsupported:
                raise errors.WordSearchError(f"Unexpected empty output_report after calling GwWordSearch\n\toutput_report: {output_report}")

        return gw_return_object

    @glasswall.utils.deprecated_alias(xml_config="content_management_policy")
    def redact_directory(self, input_directory: str, content_management_policy: Union[str, bytes, bytearray, io.BytesIO, glasswall.content_management.policies.policy.Policy], output_directory: Optional[str] = None, output_report_directory: Optional[str] = None, homoglyphs: Union[None, str, bytes, bytearray, io.BytesIO] = None, raise_unsupported: bool = True):
        """ Redacts all files in a directory and it's subdirectories using the given content_management_policy and homoglyphs file. The redacted files are written to output_directory maintaining the same directory structure as input_directory.

        Args:
            input_directory (str): The input directory containing files to redact.
            output_directory (str): The output directory where the redacted files will be written.
            output_report_directory (Optional[str], optional): Default None. If str, the output directory where analysis reports for each redacted file will be written.
            content_management_policy (Union[str, bytes, bytearray, io.BytesIO)]): The content management policy to apply.
            homoglyphs (Union[None, str, bytes, bytearray, io.BytesIO)], optional): Default None. The homoglyphs file path, str, or bytes.
            raise_unsupported (bool, optional): Default True. Raise exceptions when Glasswall encounters an error. Fail silently if False.

        Returns:
            redacted_files_dict (dict): A dictionary of file paths relative to input_directory, and glasswall.GwReturnObj with attributes: "status" (int), "output_file" (bytes), "output_report" (bytes)
        """
        redacted_files_dict = {}
        # Call redact_file on each file in input_directory
        for input_file in utils.list_file_paths(input_directory):
            relative_path = os.path.relpath(input_file, input_directory)
            # Construct paths for output file and output report
            output_file = None if output_directory is None else os.path.join(os.path.abspath(output_directory), relative_path)
            output_report = None if output_report_directory is None else os.path.join(os.path.abspath(output_report_directory), relative_path + ".xml")

            result = self.redact_file(
                input_file=input_file,
                output_file=output_file,
                output_report=output_report,
                homoglyphs=homoglyphs,
                content_management_policy=content_management_policy,
                raise_unsupported=raise_unsupported,
            )

            redacted_files_dict[relative_path] = result

        return redacted_files_dict





class FileTypeEnum:
    """ Base class for all file type enumerations. """
    pass


class FileTypeEnumError(FileTypeEnum, Exception):
    """ Enumerations that correspond to an error. """
    pass


class FileTypeEnumSuccess(FileTypeEnum):
    """ Enumerations that correspond to a file type. """
    pass




from glasswall.libraries.word_search.classes import WordSearchError


class UnknownErrorCode(WordSearchError):
    """ Unknown error code. """
    pass


# Statuses from sdk.word.search\src\glasswall.word.search\code\common\common.def.h
class Fail(WordSearchError):
    """ WordSearch error code 0. """
    pass


error_codes = {
    0: Fail,
}



from multiprocessing import Queue
from typing import Any, Callable, Optional, Union
from glasswall.multiprocessing.deletion import force_object_under_target_size


class Task:
    def __init__(
        self,
        func: Callable,
        args: Optional[tuple] = None,
        kwargs: Optional[dict] = None,
    ):
        self.func = func
        self.args = args or tuple()
        self.kwargs = kwargs or dict()

    def __eq__(self, other):
        if isinstance(other, Task):
            return (self.func, self.args, self.kwargs) == (other.func, other.args, other.kwargs)
        return False

    def __hash__(self):
        kwargs_tuple = tuple(sorted(self.kwargs.items()))
        return hash((self.func, self.args, kwargs_tuple))

    def __repr__(self):
        args_str = ", ".join(map(repr, self.args))
        kwargs_str = ", ".join(f"{key}={value!r}" for key, value in self.kwargs.items())
        return (
            f"Task(func={self.func.__name__}, args=({args_str}), kwargs=({kwargs_str}))"
        )


class TaskResult:
    timeout_seconds: Optional[float]
    memory_limit_in_gb: Optional[float]
    start_time: float
    end_time: float
    elapsed_time: float
    out_of_memory: bool
    timed_out: bool

    def __init__(
        self,
        task: Task,
        success: bool,
        result: Any = None,
        exception: Union[Exception, None] = None,
    ):
        self.task = task
        self.success = success
        self.result = result
        self.exception = exception

    def __eq__(self, other):
        if isinstance(other, TaskResult):
            return (self.task, self.success, self.result, self.exception) == \
                (other.task, other.success, other.result, other.exception)
        return False

    def __hash__(self):
        return hash((self.task, self.success, self.result, self.exception))


def execute_task_and_put_in_queue(task: Task, queue: "Queue[TaskResult]") -> None:
    try:
        func_result = task.func(*task.args, **task.kwargs)
        task_result = TaskResult(task=task, success=True, result=func_result)
    except Exception as e:
        task_result = TaskResult(task=task, success=False, exception=e)

    # Ensure task_result can fit in queue
    task_result = force_object_under_target_size(task_result, target_size=8192)

    queue.put(task_result)





import unittest
from multiprocessing import Queue

from glasswall.multiprocessing.deletion import force_object_under_target_size, Deleted
from glasswall.multiprocessing.tasks import Task, TaskResult, execute_task_and_put_in_queue


def sample_function(x, n=2):
    return x * n


class TestTasks(unittest.TestCase):
    def test_task_creation(self):
        # Test task creation
        def sample_task():
            return "Hello, world!"

        task = Task(sample_task)
        self.assertEqual(task.func, sample_task)
        self.assertEqual(task.args, ())
        self.assertEqual(task.kwargs, {})

    def test_task_creation_with_args_and_kwargs(self):
        # Test task creation with arguments and keyword arguments
        def sample_task_with_args(arg1, arg2, kwarg1=None):
            return f"Arguments: {arg1}, {arg2}. Keyword Argument: {kwarg1}"

        args = (10, 20)
        kwargs = {"kwarg1": "test"}
        task = Task(sample_task_with_args, args=args, kwargs=kwargs)
        self.assertEqual(task.func, sample_task_with_args)
        self.assertEqual(task.args, args)
        self.assertEqual(task.kwargs, kwargs)

    def test_task_result_creation(self):
        # Test task result creation
        task = Task(sample_function, args=(5,))
        task_result = TaskResult(task, success=True, result=10)
        self.assertEqual(task_result.task, task)
        self.assertTrue(task_result.success)
        self.assertEqual(task_result.result, 10)
        self.assertIsNone(task_result.exception)

    def test_execute_task_and_put_in_queue(self):
        # Test executing a task and putting it into the queue
        queue: "Queue[TaskResult]" = Queue()
        task = Task(sample_function, args=(5,))
        execute_task_and_put_in_queue(task, queue)
        task_result = queue.get()
        self.assertTrue(task_result.success)
        self.assertEqual(task_result.result, 10)

        # After getting from the queue, it should be empty
        self.assertTrue(queue.empty())

        # Additional check to ensure the task result matches the expected task
        self.assertEqual(task_result.task, task)

    def test_delete_objects_result_exceeds_threshold(self):
        # Test case where task_result size exceeds the threshold
        task_result = TaskResult(
            task=None,
            success=True,
            result="A" * 8193,
        )

        # Delete objects until total size is <= 8192 bytes
        task_result = force_object_under_target_size(task_result, target_size=8192)

        # Assert that the result is an instance of Delete
        self.assertIsInstance(task_result.result, Deleted)

    def test_delete_objects_task_kwarg_exceeds_threshold(self):
        # Test case where task_result size exceeds the threshold
        task_result = TaskResult(
            task=Task(
                func=None,
                args=None,
                kwargs={"somekwarg": "A" * 8193}
            ),
            success=True,
            result="A" * 8193,
        )

        # Delete objects until total size is <= 8192 bytes
        task_result = force_object_under_target_size(task_result, target_size=8192)

        # Assert that the Task is an instance of Delete
        self.assertIsInstance(task_result.task, Deleted)

    def test_queue_task_result_exceeds_threshold(self):
        # Test case where task_result size exceeds the threshold
        # sample_function returns "A" * 8193
        task = Task(func=sample_function, args=("A",), kwargs={"n": 8193})
        queue: "Queue[TaskResult]" = Queue()

        # Execute the function
        execute_task_and_put_in_queue(task, queue)
        task_result = queue.get()

        # Assert that the result is an instance of Delete
        self.assertIsInstance(task_result.result, Deleted)


if __name__ == "__main__":
    unittest.main()





import glasswall
from glasswall.content_management.config_elements.config_element import ConfigElement
from glasswall.content_management.policies.policy import Policy
from glasswall.content_management.switches.switch import Switch


class WordSearch(Policy):
    """ A content management policy for Word Search.

    WordSearch(default="allow", config={
        "textSearchConfig": {
            "@libVersion": "core2",
            "textList": [
                {"name": "textItem", "switches": [
                    {"name": "text", "value": "generic"},
                    {"name": "textSetting", "@replacementChar": "*", "value": "redact"},
                ]},
            ]
        }
    })
    """

    def __init__(self, default: str = "allow", config: dict = {}):
        self.default = default
        self.default_config_elements = [
            glasswall.content_management.config_elements.pdfConfig(default=default),
            glasswall.content_management.config_elements.pptConfig(default=default),
            glasswall.content_management.config_elements.sysConfig(interchange_type="xml"),
            glasswall.content_management.config_elements.tiffConfig(default=default),
            glasswall.content_management.config_elements.wordConfig(default=default),
            glasswall.content_management.config_elements.xlsConfig(default=default),
        ]
        self.config = config

        super().__init__(
            default=self.default,
            default_config_elements=self.default_config_elements,
            config=self.config,
        )

    def add_textItem(self, text: str, replacementChar: str, textSetting: str = "redact", **kwargs):
        """ Adds a textItem to the textSearchConfig textList subelements. """
        textSearchConfig = next(iter(c for c in self.config_elements if c.name == "textSearchConfig"), None)
        if not textSearchConfig:
            # Add textSearchConfig to ConfigElements
            textSearchConfig = ConfigElement(
                name="textSearchConfig",
                attributes={"libVersion": kwargs.get("libVersion", "core2")},
                subelements=[ConfigElement(name="textList")]
            )
            self.add_config_element(
                config_element=textSearchConfig,
            )

        # Select the ConfigElement named textList
        textList = next(iter(s for s in textSearchConfig.subelements if s.name == "textList"))

        # If textList has a subelement textItem that contains a switch named "text" with the same .value as arg "text", delete it to avoid duplicates.
        #   (cannot redact "generic" with "*" and also redact "generic" with "@")
        for textItem in textList.subelements.copy():
            for switch in textItem.switches:
                if switch.name == "text" and switch.value.lower() == text.lower():
                    # Remove textItem from tetList.subelements using the builtin list .remove method
                    textList.subelements.remove(textItem)
                    break

        # Add textItem to end of textList.subelements
        textList.subelements.append(
            ConfigElement(
                name="textItem",
                switches=[
                    Switch(name="text", value=text),
                    Switch(name="textSetting", attributes={"replacementChar": replacementChar}, value=textSetting)
                ],
            )
        )

        # Don't sort textList: this preserves top-down order for redaction settings.

    def remove_textItem(self, text: str):
        """ Removes a textItem from the textSearchConfig textList subelements. """
        textSearchConfig = next(iter(c for c in self.config_elements if c.name == "textSearchConfig"), None)
        if not textSearchConfig:
            raise glasswall.content_management.errors.config_elements.ConfigElementNotFound("textSearchConfig")

        # Select the ConfigElement named textList
        textList = next(iter(s for s in textSearchConfig.subelements if s.name == "textList"))

        all_textItem_texts = [
            switch.value.lower()
            for textItem in textList.subelements
            for switch in textItem.switches
            if switch.name == "text"
        ]
        if text not in all_textItem_texts:
            raise glasswall.content_management.errors.switches.SwitchNotFound(f"No switch found with name: 'text' and value: '{text}'")

        # If textList has a subelement textItem that contains a switch named "text" with the same .value as arg "text", delete it to avoid duplicates.
        #   (cannot redact "generic" with "*" and also redact "GeNeRiC" with "@")
        for textItem in textList.subelements.copy():
            for switch in textItem.switches:
                if switch.name == "text" and switch.value.lower() == text.lower():
                    textList.subelements.remove(textItem)
                    break





from .classes import ArchiveManagerSuccess


# Statuses from sdk.ArchiveManager\src\core.support\sdk.interface\gw2_returnstatus.h
class Success(ArchiveManagerSuccess):
    """ ArchiveManager success code 1. """
    pass


success_codes = {
    1: Success,
}



import functools
import sys
from typing import Callable, TypeVar


class Deleted:
    """ Object deemed too large for queue. """
    pass


def hash_dict(func: Callable):
    """Transform mutable dictionary into immutable"""
    class HashDict(dict):
        def __hash__(self):
            return hash(frozenset(self.items()))

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        args = tuple([HashDict(arg) if isinstance(arg, dict) else arg for arg in args])
        kwargs = {k: HashDict(v) if isinstance(v, dict) else v for k, v in kwargs.items()}
        return func(*args, **kwargs)
    return wrapped


@hash_dict
@functools.lru_cache(maxsize=None)  # Cache the sizes of objects
def get_size(obj: object) -> int:
    """Recursively find the total size in memory of an object."""
    if isinstance(obj, dict):
        # Convert dictionary items to a tuple of tuples (key, value)
        return sum(get_size(k) + get_size(v) for k, v in obj.items())
    elif isinstance(obj, (list, tuple, set)):
        # Iterate over elements and sum their sizes
        return sum(get_size(x) for x in obj)
    elif isinstance(obj, (str, bytes, int, float)):
        # For simple types, use sys.getsizeof
        return sys.getsizeof(obj)
    elif hasattr(obj, "__dict__"):
        # For custom objects, iterate over their attributes
        # Convert attributes to a tuple to ensure hashability
        return sum(get_size(attr) for attr in vars(obj).values())
    else:
        return sys.getsizeof(obj)


T = TypeVar("T")


def force_object_under_target_size(obj: T, target_size: int = 8192) -> T:
    """Delete objects within the given object until total size is <= target_size"""
    total_size = get_size(obj)
    if total_size > target_size:
        attrs = sorted(obj.__dict__.items(), key=lambda x: get_size(x[1]), reverse=True)
        while total_size > target_size and attrs:
            largest_obj, largest_value = attrs.pop(0)
            setattr(obj, largest_obj, Deleted())
            total_size -= get_size(largest_value)
        if total_size > target_size:
            # Shouldn't ever happen, but if task_result is still too big ungracefully exit instead of hanging
            raise Exception(f"TaskResult too big for queue.\nSize: '{get_size(obj)}'.\nDict: '{obj.__dict__}'")
    return obj





from glasswall.libraries.editor.classes import EditorSuccess


# Statuses from sdk.editor\src\core.support\sdk.interface\gw2_returnstatus.h
class OK(EditorSuccess):
    """ Editor success code 0. """
    pass


class OKWithCleaning(EditorSuccess):
    """ Editor success code 1. """
    pass


success_codes = {
    0: OK,
    1: OKWithCleaning,
}



import argparse
import os
import sys
import time
import unittest
from collections import Counter

from tqdm import tqdm

import glasswall
from glasswall.config.logging import log
from glasswall.multiprocessing import GlasswallProcessManager, Task

glasswall.config.logging.console.setLevel("CRITICAL")


def worker_function(*args, **kwargs):
    library_name = os.environ.get("GLASSWALL_LIBRARY_NAME")
    function_name = os.environ.get("GLASSWALL_FUNCTION_NAME")
    library_path = os.environ.get("GLASSWALL_LIBRARY_PATH")
    library_class = getattr(glasswall, library_name)
    library = library_class(library_path)
    library_function = getattr(library, function_name)
    library_function(*args, **kwargs)


class TestIntegration(unittest.TestCase):
    input_directory: str
    output_directory: str
    library_directory: str
    max_workers: int
    timeout: float
    library_name: str
    function_name: str

    @classmethod
    def setUpClass(cls):
        cls.input_directory = os.environ["GLASSWALL_INPUT_PATH"]
        cls.output_directory = os.environ["GLASSWALL_OUTPUT_PATH"]
        cls.library_directory = os.environ["GLASSWALL_LIBRARY_PATH"]
        if "GLASSWALL_MAX_WORKERS" in os.environ.keys():
            cls.max_workers = int(os.environ["GLASSWALL_MAX_WORKERS"])
        else:
            cls.max_workers = os.cpu_count() or 1
        cls.timeout = float(os.environ["GLASSWALL_TIMEOUT"])
        cls.memory_limit_in_gb = int(os.environ["GLASSWALL_MEMORY_LIMIT_IN_GB"])
        cls.library_name = os.environ["GLASSWALL_LIBRARY_NAME"]
        cls.function_name = os.environ["GLASSWALL_FUNCTION_NAME"]
        cls.multiply_dataset = int(os.environ["GLASSWALL_MULTIPLY_DATASET"])

        print(f"input_directory:{cls.input_directory}")
        print(f"output_directory:{cls.output_directory}")
        print(f"library_directory:{cls.library_directory}")
        print(f"max_workers:{cls.max_workers}")
        print(f"timeout:{cls.timeout}")
        print(f"library_name:{cls.library_name}")
        print(f"function_name:{cls.function_name}")

        if not os.path.isdir(cls.input_directory):
            raise NotADirectoryError(cls.input_directory)

        if not os.path.isdir(cls.output_directory):
            os.makedirs(cls.output_directory, exist_ok=True)

        if not os.path.isdir(cls.library_directory) and not os.path.isfile(
            cls.library_directory
        ):
            raise NotADirectoryError(cls.library_directory)

        if not hasattr(glasswall.Editor, cls.function_name):
            raise AttributeError(cls.function_name)

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_integration(self):
        input_files = glasswall.utils.list_file_paths(self.input_directory)
        if self.multiply_dataset:
            input_files *= self.multiply_dataset

        start_time = time.time()
        with GlasswallProcessManager(
            max_workers=self.max_workers,
            worker_timeout_seconds=self.timeout,
            memory_limit_in_gb=self.memory_limit_in_gb,
        ) as process_manager:
            for input_file in tqdm(input_files, desc="Queueing files"):
                relative_path = os.path.relpath(input_file, self.input_directory)
                output_file = os.path.normpath(
                    os.path.join(self.output_directory, relative_path)
                )

                task = Task(
                    func=worker_function,
                    args=None,
                    kwargs={
                        "input_file": input_file,
                        "output_file": output_file,
                    },
                )
                process_manager.queue_task(task)
        end_time = time.time()

        # Reenable logging
        glasswall.config.logging.console.setLevel("INFO")

        log.info(f"Elapsed: {end_time - start_time}")

        results = process_manager.task_results

        log.info("Exceptions:")
        exceptions = Counter(str(item.exception) if item else None for item in results)
        for k, v in exceptions.items():
            log.info(f"{k} = {v}")

        success = [item.success for item in results if item.success is True]
        failure = [item.success for item in results if item.success is False]
        log.info(f"Success: {len(success)}")
        log.info(f"Failure: {len(failure)}")

        self.assertFalse(all(result.success is False for result in results), msg="All 'success' attributes are False.")

        for task_result in results:
            log.info(task_result.__dict__)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--input-directory", required=True, type=str)
    parser.add_argument("--output-directory", required=True, type=str)
    parser.add_argument("--library-directory", required=True, type=str)
    parser.add_argument("--max-workers", required=True, type=int)
    parser.add_argument("--timeout", required=True, type=float)
    parser.add_argument("--memory-limit-in-gb", default="0", type=int)
    parser.add_argument("--library-name", default="Editor", type=str)
    parser.add_argument("--function-name", default="export_file", type=str)
    parser.add_argument("--multiply-dataset", default="1", type=int)
    args = parser.parse_args()

    os.environ["GLASSWALL_INPUT_PATH"] = str(args.input_directory)
    os.environ["GLASSWALL_OUTPUT_PATH"] = str(args.output_directory)
    os.environ["GLASSWALL_LIBRARY_PATH"] = str(args.library_directory)
    os.environ["GLASSWALL_MAX_WORKERS"] = str(args.max_workers)
    os.environ["GLASSWALL_TIMEOUT"] = str(args.timeout)
    os.environ["GLASSWALL_MEMORY_LIMIT_IN_GB"] = str(args.memory_limit_in_gb)
    os.environ["GLASSWALL_LIBRARY_NAME"] = str(args.library_name)
    os.environ["GLASSWALL_FUNCTION_NAME"] = str(args.function_name)
    os.environ["GLASSWALL_MULTIPLY_DATASET"] = str(args.multiply_dataset)

    # Set environment variables for worker_function to avoid below error via subclassing unittest.TestCase
    # TypeError: cannot pickle '_io.TextIOWrapper' object

    # unittest needs a clean sys.argv
    while len(sys.argv) > 1:
        sys.argv.pop()

    unittest.main()




from glasswall.libraries.security_tagging.classes import SecurityTaggingError


class UnknownErrorCode(SecurityTaggingError):
    """ Unknown error code. """
    pass


# TODO expand, pending security tagging status code enhancements
class GeneralFail(SecurityTaggingError):
    """ SecurityTagging error code 0. """
    pass


error_codes = {
    0: GeneralFail,
}





from glasswall.determine_file_type.classes import FileTypeEnumError


class UnknownErrorCode(FileTypeEnumError):
    """ Unknown error code. """
    pass


# The default when we don't know or can't determine the file type.
class ft_unknown(FileTypeEnumError):
    integer = 0
    string = "unknown"


# Not related to filetypes, but these are required since they give information when something goes wrong.
class ft_fileIssues(FileTypeEnumError):
    integer = 1
    string = "fileIssues"


class ft_bufferIssues(FileTypeEnumError):
    integer = 2
    string = "bufferIssues"


class ft_internalIssues(FileTypeEnumError):
    integer = 3
    string = "internalIssues"


class ft_licenseExpired(FileTypeEnumError):
    integer = 4
    string = "licenseExpired"


class ft_passwordProtectedOpcFile(FileTypeEnumError):
    integer = 5
    string = "passwordProtectedOpcFile"


class ft_nullPointerArgument(FileTypeEnumError):
    integer = 6
    string = "nullPointerArgument"





# Note, `dependencies` refers to the dependencies required to call
# ctypes.cdll.LoadLibrary on CentOS7 / Windows 10 and may not be indicative
# of the full dependencies for all functionality within a library.

os_info = {
    "Linux": {
        "archive_manager": {
            "file_name": "libglasswall.archive.manager.so",
            "dependencies": [],
            "relies_on": ["editor"],
        },
        "editor": {
            "file_name": "libglasswall_core2.so",
            "dependencies": [],
        },
        "editor_cli": {
            "file_name": "gwcli",
            "dependencies": [],
            "relies_on": ["editor"],
        },
        "rebuild": {
            "file_name": "libglasswall.classic.so",
            "dependencies": [],
        },
        "security_tagging": {
            "file_name": "libgw_securtag.so",
            "dependencies": [
                "libicudata.so.56",
                "libicui18n.so.56",
                "libicuuc.so.56",
                "libQt5Core.so.5",
                "libQt5Multimedia.so.5",
                "libQt5XmlPatterns.so.5"
            ],
            "relies_on": ["editor"],
        },
        "word_search": {
            "file_name": "libglasswall.word.search.so",
            # loading dependencies through ctypes doesn't help for word search, use LD_LIBRARY_PATH
            "dependencies": [],
            "relies_on": ["editor"],
        },
    },

    "Windows": {
        "archive_manager": {
            "file_name": "glasswall.archive.manager.dll",
            "dependencies": [],
            "relies_on": ["editor"],
        },
        "editor": {
            "file_name": "glasswall_core2.dll",
            "dependencies": [
                "Qt5Core.dll",
                "Qt5Xml.dll"
            ],
        },
        "editor_cli": {
            "file_name": "gwcli.exe",
            "dependencies": [],
            "relies_on": ["editor"],
        },
        "rebuild": {
            "file_name": "glasswall.classic.dll",
            "dependencies": [],
        },
        "security_tagging": {
            "file_name": "gw_securtag.dll",
            "dependencies": [
                "Qt5Core.dll",
                "Qt5Gui.dll",
                "Qt5Xml.dll",
                "Qt5XmlPatterns.dll",
                "Qt5Network.dll"
            ],
            "relies_on": ["editor"],
        },
        "word_search": {
            "file_name": "glasswall.word.search.dll",
            "dependencies": [],
            "relies_on": ["editor"],
        },
    },
}





class ArchiveManagerError(Exception):
    """ Base class for all Glasswall ArchiveManager errors. """
    pass


class ArchiveManagerSuccess:
    """ Base class for all Glasswall ArchiveManager successes. """
    pass





from glasswall.content_management.errors.policies import ContentManagementPolicyError


class SwitchNotFound(ContentManagementPolicyError):
    """ The Content Management Policy Switch could not be found. """
    pass


class RestrictedValue(ContentManagementPolicyError):
    """ The Content Management Policy Switch has an unexpected value. """
    pass



from multiprocessing import Process, Queue
import os
import time
from collections import deque
from typing import List, Optional

from tqdm import tqdm

from glasswall.multiprocessing.task_watcher import TaskWatcher
from glasswall.multiprocessing.tasks import Task, TaskResult


class GlasswallProcessManager:
    def __init__(
        self,
        max_workers: Optional[int] = None,
        worker_timeout_seconds: Optional[float] = None,
        memory_limit_in_gb: Optional[float] = None,
        sleep_time: Optional[float] = None,
    ):
        self.max_workers = max_workers or os.cpu_count() or 1
        self.worker_timeout_seconds = worker_timeout_seconds
        self.memory_limit_in_gb = memory_limit_in_gb
        self.sleep_time = sleep_time

        self.pending_processes: "deque[Process]" = deque()
        self.active_processes: list[Process] = []
        self.task_results_queue: "Queue[TaskResult]" = Queue()
        self.task_results: List[TaskResult] = []

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.start_tasks()

    def queue_task(self, task: Task):
        # Create and queue the process without starting it
        process = Process(
            target=TaskWatcher,
            kwargs={
                "task": task,
                "task_results_queue": self.task_results_queue,
                "timeout_seconds": self.worker_timeout_seconds,
                "memory_limit_in_gb": self.memory_limit_in_gb,
            },
        )
        self.pending_processes.append(process)

    def clean_task_results_queue(self):
        while not self.task_results_queue.empty():
            self.task_results.append(self.task_results_queue.get())

    def remove_completed_active_processes(self, pbar: tqdm):
        index = 0
        while index < len(self.active_processes):
            process = self.active_processes[index]
            if not process.is_alive():
                process.join()
                self.active_processes.remove(process)
                pbar.update(1)
                self.clean_task_results_queue()
            else:
                index += 1

    def wait_for_completed_process(self, pbar: tqdm):
        self.remove_completed_active_processes(pbar)
        while len(self.active_processes) >= self.max_workers:
            self.remove_completed_active_processes(pbar)
            if self.sleep_time:
                time.sleep(self.sleep_time)

    def start_tasks(self):
        with tqdm(total=len(self.pending_processes), desc="Processing tasks") as pbar:
            while self.pending_processes or len(self.active_processes) > 0:
                if self.active_processes:
                    self.wait_for_completed_process(pbar)

                while len(self.active_processes) < self.max_workers and self.pending_processes:
                    process = self.pending_processes.popleft()
                    self.active_processes.append(process)
                    process.start()





from glasswall.content_management.config_elements.config_element import ConfigElement


class textList(ConfigElement):
    """ A textList ConfigElement. """

    def __init__(self, subelements: list = []):
        self.name = self.__class__.__name__
        self.subelements = subelements
        super().__init__(name=self.name, subelements=self.subelements)





import os
import subprocess
import unittest


def extract_python_blocks_from_readme(file_path):
    # Function to extract Python blocks from the README and return as a dictionary
    python_blocks = {}
    start_marker = "```py"   # Start marker for identifying Python code blocks in the README
    end_marker = "```"       # End marker for identifying Python code blocks in the README
    header_count = {}        # Dictionary to keep track of header occurrences

    # Read the content of the README.md file
    with open(file_path, "rb") as readme_file:
        readme_content = readme_file.read()

    readme_content = readme_content.decode("utf-8")

    start_index = 0
    while start_index != -1:
        # Find the next occurrence of the start_marker
        start_index = readme_content.find(start_marker, start_index)
        if start_index != -1:
            # Find the corresponding end index of the current Python block
            end_index = readme_content.find(end_marker, start_index + len(start_marker))
            if end_index != -1:
                # Find the header just before the current Python block
                header_start = readme_content.rfind("###", 0, start_index)
                header_end = readme_content.find("\n", header_start)
                header = readme_content[header_start:header_end].strip("#").strip()

                # Increment the count for the current header in the dictionary
                if header not in header_count:
                    header_count[header] = 1
                else:
                    header_count[header] += 1

                # Modify the header name to include a numbering system for subsequent blocks
                header_name = f"{header}" if header_count[header] == 1 else f"{header} ({header_count[header]})"
                # Extract the Python block and add it to the dictionary with the modified header name
                python_blocks[header_name] = readme_content[start_index + len(start_marker):end_index].strip()

            # Move the start_index to the next character after the current Python block
            start_index = end_index + len(end_marker)

    return python_blocks


def execute_python_block(python_code):
    try:
        # Execute the Python block in a separate process and capture the output
        return subprocess.run(["python", "-c", python_code], stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)

    except subprocess.CalledProcessError as e:
        print(f"Execution failed with error: {e}")

        # If an error occurs, print the captured output (stdout and stderr)
        if e.stdout:
            print(e.stdout)
        if e.stderr:
            print(e.stderr)

        raise


def count_python_blocks_in_readme(file_path):
    # Function to count the occurrences of "```py" in the README
    with open(file_path, "rb") as readme_file:
        readme_content = readme_file.read()

    readme_content = readme_content.decode("utf-8")

    return readme_content.count("```py")


class TestReadmeCode(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Extract Python blocks from the README and store them in a class-level variable
        readme_file_path = "README.md"
        cls.python_blocks_dict = extract_python_blocks_from_readme(readme_file_path)
        cls.python_blocks_count = count_python_blocks_in_readme(readme_file_path)

        if not os.path.exists("C:/gwpw/libraries"):
            raise unittest.SkipTest("No Glasswall libraries found.")

    def test_extract_python_blocks(self):
        # Test that the extract_python_blocks_from_readme function works as expected
        self.assertEqual(len(self.python_blocks_dict), self.python_blocks_count)

    def test_python_blocks_execution(self):
        # Test the execution of each Python block and assert that they pass successfully
        for header, block in self.python_blocks_dict.items():
            with self.subTest(header=header):
                print(f"Running subtest for header: '{header}'")
                result = execute_python_block(block)
                if isinstance(result, subprocess.CalledProcessError):
                    self.fail(f"Test '{header}' failed with error: {result}")
                else:
                    self.assertEqual(result.returncode, 0, f"Test '{header}' failed with a non-zero exit code.")


if __name__ == "__main__":
    unittest.main()





import time
import unittest
from multiprocessing import Process

from glasswall.multiprocessing.manager import GlasswallProcessManager
from glasswall.multiprocessing.tasks import Task, TaskResult


def sample_task():
    return "Task completed!"


def exception_task():
    raise ValueError("Test exception")


class TestGlasswallProcessManager(unittest.TestCase):
    def test_queue_task(self):
        # Test queuing a task
        manager = GlasswallProcessManager(max_workers=2)
        self.assertEqual(len(manager.pending_processes), 0)
        task = Task(sample_task)
        manager.queue_task(task)
        self.assertEqual(len(manager.pending_processes), 1)
        pending_process = manager.pending_processes[0]
        self.assertIsInstance(pending_process, Process)
        self.assertEqual(task, pending_process._kwargs["task"])

    def test_start_tasks(self):
        # Test starting tasks
        manager = GlasswallProcessManager(max_workers=2)
        task = Task(sample_task)
        manager.queue_task(task)
        self.assertEqual(len(manager.pending_processes), 1)
        self.assertEqual(len(manager.active_processes), 0)
        manager.start_tasks()
        # start_tasks populates active_processes, but should be empty once tasks are done
        self.assertEqual(len(manager.pending_processes), 0)
        self.assertEqual(len(manager.active_processes), 0)

    def test_clean_task_results_queue(self):
        # Test cleaning task results queue
        manager = GlasswallProcessManager(max_workers=2)
        task1 = Task(sample_task)
        task2 = Task(sample_task)
        result1 = TaskResult(task1, success=True, result="Task 1 completed!")
        result2 = TaskResult(task2, success=True, result="Task 2 completed!")

        # Insert some task results into the queue
        manager.task_results_queue.put(result1)
        manager.task_results_queue.put(result2)

        # Introduce a small delay to prevent race condition
        time.sleep(0.1)

        # Ensure the queue is not empty before cleaning
        self.assertFalse(manager.task_results_queue.empty())

        # Clean the task results queue
        manager.clean_task_results_queue()

        # Ensure the queue is empty after cleaning
        self.assertTrue(manager.task_results_queue.empty())

        # Ensure task results are correctly appended to manager's task_results list
        self.assertEqual(len(manager.task_results), 2)


if __name__ == "__main__":
    unittest.main()





import os
import setuptools

current_directory = os.path.dirname(os.path.abspath(__file__))
readme_path = os.path.join(current_directory, "README.md")
with open(readme_path, "r", encoding="utf-8") as f:
    long_description = f.read()


setuptools.setup(
    name="glasswall",
    description="Glasswall Python Wrapper",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="AngusWR",
    author_email="aroberts@glasswall.com",
    url="https://github.com/gw-engineering/glasswall-python-wrapper",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.6",
    ],
    python_requires=">=3.6",
    include_package_data=True,
    install_requires=[
        "lxml>=4.9.4,<5",
        "psutil>=5.9.8,<6",
        "tqdm>=4.64.1,<5",
    ],
)





import time
import unittest
from multiprocessing import Queue

from glasswall.multiprocessing.task_watcher import TaskWatcher
from glasswall.multiprocessing.tasks import Task, TaskResult


def sample_task() -> str:
    return "Task completed!"


def long_running_task() -> None:
    time.sleep(1)


def exception_task() -> None:
    raise ValueError("Test exception")


class TestTaskWatcher(unittest.TestCase):
    def test_task_watcher_auto_start(self):
        # Test task watcher with auto_start=True
        task = Task(sample_task)
        queue: "Queue[TaskResult]" = Queue()
        TaskWatcher(task, queue)
        task_result = queue.get()
        self.assertTrue(task_result.success)
        self.assertEqual(task_result.result, "Task completed!")

    def test_task_watcher_manual_start(self):
        # Test task watcher with auto_start=False
        task = Task(sample_task)
        queue: "Queue[TaskResult]" = Queue()
        watcher = TaskWatcher(task, queue, auto_start=False)
        watcher.start_task()
        watcher.watch_task()
        watcher.update_queue()
        task_result = queue.get()
        self.assertTrue(task_result.success)
        self.assertEqual(task_result.result, "Task completed!")

    def test_task_watcher_timeout(self):
        # Test task watcher with timeout
        task = Task(long_running_task)
        queue: "Queue[TaskResult]" = Queue()
        TaskWatcher(task, queue, timeout_seconds=0.1)
        task_result = queue.get()
        self.assertFalse(task_result.success)
        self.assertIsInstance(task_result.exception, TimeoutError)

    def test_task_watcher_exception(self):
        # Test task watcher with task raising an exception
        task = Task(exception_task)
        queue: "Queue[TaskResult]" = Queue()
        TaskWatcher(task, queue)
        task_result = queue.get()
        self.assertFalse(task_result.success)
        self.assertIsInstance(task_result.exception, ValueError)
        self.assertIn("Test exception", str(task_result.exception))


if __name__ == "__main__":
    unittest.main()





from glasswall.content_management.switches.switch import Switch


class acroform(Switch):
    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["allow", "disallow", "sanitise"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class actions_all(Switch):
    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["allow", "disallow", "sanitise"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class digital_signatures(Switch):
    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["allow", "disallow", "sanitise"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class embedded_files(Switch):
    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["allow", "disallow", "sanitise"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class embedded_images(Switch):
    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["allow", "disallow", "sanitise"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class external_hyperlinks(Switch):
    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["allow", "disallow", "sanitise"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class internal_hyperlinks(Switch):
    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["allow", "disallow", "sanitise"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class javascript(Switch):
    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["allow", "disallow", "sanitise"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class metadata(Switch):
    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["allow", "disallow", "sanitise"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class value_outside_reasonable_limits(Switch):
    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["allow", "disallow", "sanitise"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class watermark(Switch):
    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = None
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)





from .classes import ArchiveManagerError


class UnknownErrorCode(ArchiveManagerError):
    """ Unknown error code. """
    pass


# Statuses from sdk.archive.manager\src\glasswall.archive\code\common\custom.types.h
class Fail(ArchiveManagerError):
    """ ArchiveManager error code 0. """
    pass


error_codes = {
    0: Fail,
}





from glasswall.content_management.switches.switch import Switch


class bmp(Switch):
    """ An ArchiveManager bmp switch. """

    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["no_action", "discard", "process"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class doc(Switch):
    """ An ArchiveManager doc switch. """

    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["no_action", "discard", "process"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class docx(Switch):
    """ An ArchiveManager docx switch. """

    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["no_action", "discard", "process"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class elf(Switch):
    """ An ArchiveManager elf switch. """

    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["no_action", "discard", "process"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class emf(Switch):
    """ An ArchiveManager emf switch. """

    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["no_action", "discard", "process"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class gif(Switch):
    """ An ArchiveManager gif switch. """

    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["no_action", "discard", "process"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class jpeg(Switch):
    """ An ArchiveManager jpeg switch. """

    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["no_action", "discard", "process"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class mp3(Switch):
    """ An ArchiveManager mp3 switch. """

    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["no_action", "discard", "process"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class mp4(Switch):
    """ An ArchiveManager mp4 switch. """

    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["no_action", "discard", "process"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class mpg(Switch):
    """ An ArchiveManager mpg switch. """

    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["no_action", "discard", "process"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class o(Switch):
    """ An ArchiveManager o switch. """

    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["no_action", "discard", "process"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class pdf(Switch):
    """ An ArchiveManager pdf switch. """

    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["no_action", "discard", "process"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class pe(Switch):
    """ An ArchiveManager pe switch. """

    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["no_action", "discard", "process"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class png(Switch):
    """ An ArchiveManager png switch. """

    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["no_action", "discard", "process"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class ppt(Switch):
    """ An ArchiveManager ppt switch. """

    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["no_action", "discard", "process"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class pptx(Switch):
    """ An ArchiveManager pptx switch. """

    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["no_action", "discard", "process"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class tiff(Switch):
    """ An ArchiveManager tiff switch. """

    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["no_action", "discard", "process"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class txt(Switch):
    """ An ArchiveManager txt switch. """

    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["no_action", "discard", "process"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class wav(Switch):
    """ An ArchiveManager wav switch. """

    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["no_action", "discard", "process"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class wmf(Switch):
    """ An ArchiveManager wmf switch. """

    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["no_action", "discard", "process"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class xls(Switch):
    """ An ArchiveManager xls switch. """

    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["no_action", "discard", "process"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)


class xlsx(Switch):
    """ An ArchiveManager xlsx switch. """

    def __init__(self, value: str):
        self.name = self.__class__.__name__
        self.restrict_values = ["no_action", "discard", "process"]
        self.value = value
        super().__init__(name=self.name, value=self.value, restrict_values=self.restrict_values)
